import copy

import numpy as np
import pandas as pd
import pytest

import pharmpy.data as data
from pharmpy import Model


@pytest.fixture
def df():
    df = data.PharmDataFrame(
        {
            'ID': [1, 1, 2, 2],
            'DV': [0.1, 0.2, 0.5, 0.6],
            'WGT': [70, 72, 75, 75],
            'HGT': [185, 185, 160, 160],
        }
    )
    df.pharmpy.column_type[('ID', 'DV', 'WGT', 'HGT')] = [
        data.ColumnType.ID,
        data.ColumnType.DV,
        data.ColumnType.COVARIATE,
        data.ColumnType.COVARIATE,
    ]
    return df


@pytest.fixture
def df2():
    df = data.PharmDataFrame(
        {
            'ID': [1, 1, 2, 2],
            'DV': [0, 0.2, 0, 0.6],
            'WGT': [70, 72, 75, 75],
            'HGT': [185, 185, 160, 160],
            'TIME': [0, 1, 0, 1],
            'AMT': [1, 0, 1, 0],
        }
    )
    df.pharmpy.column_type[('ID', 'DV', 'WGT', 'HGT', 'TIME', 'AMT')] = [
        data.ColumnType.ID,
        data.ColumnType.DV,
        data.ColumnType.COVARIATE,
        data.ColumnType.COVARIATE,
        data.ColumnType.IDV,
        data.ColumnType.DOSE,
    ]
    return df


@pytest.fixture
def df3():
    df = data.PharmDataFrame(
        {
            'ID': [1, 1, 1, 1, 2, 2, 2, 2],
            'DV': [0, 0.1, 0.5, 0.5, 0, 0.3, 4, 1],
            'TIME': [0, 2, 5, 9, 0, 2, 5, 9],
            'AMT': [4, 0, 3, 0, 2, 0, 0, 0],
        }
    )
    df.pharmpy.column_type[('ID', 'DV', 'TIME', 'AMT')] = [
        data.ColumnType.ID,
        data.ColumnType.DV,
        data.ColumnType.IDV,
        data.ColumnType.DOSE,
    ]
    return df


def test_column_type():
    col_type = data.ColumnType.ID
    assert col_type == data.ColumnType.ID
    assert 'ID' in repr(col_type)
    assert col_type.max_one
    ct2 = data.ColumnType.UNKNOWN
    assert ct2 == data.ColumnType.UNKNOWN
    assert 'UNKNOWN' in repr(ct2)
    assert not ct2.max_one


def test_data_frame():
    df = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6]})
    assert list(df.columns) == ['ID', 'DV']


def test_accessor_get_set_column_type():
    df = data.PharmDataFrame(
        {
            'ID': [1, 1, 2, 2],
            'DV': [0.1, 0.2, 0.5, 0.6],
            'WGT': [70, 70, 75, 75],
            'HGT': [185, 185, 160, 160],
        }
    )
    assert df.pharmpy.column_type['ID'] == data.ColumnType.UNKNOWN
    assert df.pharmpy.column_type[('ID', 'DV')] == [
        data.ColumnType.UNKNOWN,
        data.ColumnType.UNKNOWN,
    ]

    df.pharmpy.column_type['ID'] = data.ColumnType.ID
    df.pharmpy.column_type['DV'] = data.ColumnType.DV
    with pytest.raises(KeyError):  # Max one id column
        df.pharmpy.column_type['DV'] = data.ColumnType.ID
    with pytest.raises(KeyError):
        df.pharmpy.column_type['NOEXISTS'] = data.ColumnType.COVARIATE
    with pytest.raises(ValueError):
        df.pharmpy.column_type[('ID', 'DV')] = [
            data.ColumnType.COVARIATE,
            data.ColumnType.COVARIATE,
            data.ColumnType.COVARIATE,
        ]
    with pytest.raises(KeyError):
        df.pharmpy.column_type['NOAVAIL']
    assert df.pharmpy.column_type['ID'] == data.ColumnType.ID
    assert df.pharmpy.column_type['DV'] == data.ColumnType.DV
    assert df.pharmpy.column_type[['ID', 'DV']] == [data.ColumnType.ID, data.ColumnType.DV]
    df.pharmpy.column_type[['HGT', 'WGT']] = data.ColumnType.COVARIATE
    assert df.pharmpy.labels_by_type[data.ColumnType.COVARIATE] == ['WGT', 'HGT']
    df2 = df.copy()
    assert df2.pharmpy.column_type['ID'] == data.ColumnType.ID
    assert df2.pharmpy.column_type['DV'] == data.ColumnType.DV

    assert df2.pharmpy.labels_by_type[data.ColumnType.ID] == ['ID']
    assert df2.pharmpy.labels_by_type[[data.ColumnType.ID]] == ['ID']
    assert df2.pharmpy.labels_by_type[[]] == []
    assert df2.pharmpy.labels_by_type[data.ColumnType.DV] == ['DV']
    assert df2.pharmpy.labels_by_type[[data.ColumnType.ID, data.ColumnType.DV]] == ['ID', 'DV']

    assert df2.pharmpy.id_label == 'ID'

    df2.pharmpy.column_type[['ID', 'DV']] = (data.ColumnType.COVARIATE, data.ColumnType.IDV)
    assert df2.pharmpy.column_type[['ID', 'DV']] == [data.ColumnType.COVARIATE, data.ColumnType.IDV]
    assert df.pharmpy.column_type['ID'] == data.ColumnType.ID

    df3 = data.PharmDataFrame(
        {1: [1, 1, 2, 2], 2: [0.1, 0.2, 0.5, 0.6], 3: [70, 70, 75, 75], 4: [185, 185, 160, 160]}
    )
    assert df3.pharmpy.column_type[1] == data.ColumnType.UNKNOWN
    df3.pharmpy.column_type[[3, 4]] = data.ColumnType.COVARIATE
    df3.pharmpy.column_type[1] = data.ColumnType.ID
    assert df3.pharmpy.column_type[[1, 2, 3, 4]] == [
        data.ColumnType.ID,
        data.ColumnType.UNKNOWN,
        data.ColumnType.COVARIATE,
        data.ColumnType.COVARIATE,
    ]


def test_time_varying_covariates(df):
    assert df.pharmpy.time_varying_covariates == ['WGT']
    df_untyped = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6]})
    assert df_untyped.pharmpy.time_varying_covariates == []


def test_covariate_baselines(df):
    correct_baselines = pd.DataFrame(
        {'WGT': [70, 75], 'HGT': [185, 160]}, index=pd.Int64Index([1, 2], name='ID')
    )
    pd.testing.assert_frame_equal(df.pharmpy.covariate_baselines, correct_baselines)


def test_observations(df2):
    correct_observations = (
        pd.DataFrame({'DV': [0.2, 0.6], 'ID': [1, 2], 'TIME': [1.0, 1.0]})
        .set_index(['ID', 'TIME'])
        .squeeze()
    )
    pd.testing.assert_series_equal(df2.pharmpy.observations, correct_observations)


def test_doses(df2):
    correct_doses = (
        pd.DataFrame({'AMT': [1, 1], 'ID': [1, 2], 'TIME': [0, 0]})
        .set_index(['ID', 'TIME'])
        .squeeze()
    )
    pd.testing.assert_series_equal(df2.pharmpy.doses, correct_doses)


def test_add_doseid(df2):
    df2.pharmpy.add_doseid()
    assert list(df2['DOSEID']) == [1, 1, 1, 1]


def test_add_time_after_dose(df2):
    df2.pharmpy.add_time_after_dose()
    assert list(df2['TAD']) == [0, 1, 0, 1]


def test_tad_pheno(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    print(model.dataset.dtypes)
    model.dataset.pharmpy.add_time_after_dose()


def test_concentration_parameters(df2, df3):
    df = df2.pharmpy.concentration_parameters()
    correct = pd.DataFrame(
        {
            'ID': [1, 2],
            'DOSEID': [1, 1],
            'Cmax': [0.2, 0.6],
            'Tmax': [1.0, 1.0],
            'Cmin': np.nan,
            'Tmin': np.nan,
        }
    )
    correct.set_index(['ID', 'DOSEID'], inplace=True)
    pd.testing.assert_frame_equal(df, correct)

    df = df3.pharmpy.concentration_parameters()
    correct = pd.DataFrame(
        {
            'ID': [1, 1, 2],
            'DOSEID': [1, 2, 1],
            'Cmax': [0.1, 0.5, 4.0],
            'Tmax': [2.0, 0.0, 5.0],
            'Cmin': [np.nan, 0.5, 1.0],
            'Tmin': [np.nan, 4.0, 9.0],
        }
    )
    correct.set_index(['ID', 'DOSEID'], inplace=True)
    pd.testing.assert_frame_equal(df, correct)


def test_write(fs, df):
    df.pharmpy.write_csv(path="my.csv")
    with open("my.csv", "r") as fh:
        contents = fh.read()
    assert contents == "ID,DV,WGT,HGT\n1,0.1,70,185\n1,0.2,72,185\n2,0.5,75,160\n2,0.6,75,160\n"


def test_copy(df):
    new = copy.deepcopy(df)
    assert id(new) != id(df)



import os
import subprocess
import uuid
from pathlib import Path

from pharmpy.plugins.nonmem import conf, convert_model


def execute_model(model):
    database = model.database
    model = convert_model(model)
    path = Path.cwd() / f'NONMEM_run_{model.name}-{uuid.uuid1()}'
    path.mkdir(parents=True, exist_ok=True)
    model = model.copy()
    model._dataset_updated = True  # Hack to get update_source to update IGNORE
    model.update_source(nofiles=True)
    try:
        model.dataset.name
    except AttributeError:
        model.dataset.name = "dataset"
    datapath = model.dataset.pharmpy.write_csv(path=path)
    model.dataset_path = datapath.name  # Make path in $DATA local
    model.write(path=path, force=True)
    basepath = Path(model.name)
    args = [
        nmfe_path(),
        model.name + model.filename_extension,
        str(basepath.with_suffix('.lst')),
    ]
    # Create wrapper script that cd:s into rundirectory
    # This enables the execute_model function to be parallelized
    # using threads. chdir here does not work since all threads will
    # share cwd. Using processes on Windows from R currently hangs.
    # Also the -rundir option to nmfe does not work entirely
    if os.name == 'nt':
        with open(path / 'cdwrapper.bat', 'w') as fp:
            fp.write(f"@echo off\ncd {path}\n{' '.join(args)}\n")
        cmd = str(path / 'cdwrapper.bat')
    else:
        with open(path / 'cdwrapper', 'w') as fp:
            fp.write(f"#!/bin/sh\ncd {path}\n{' '.join(args)}\n")
        os.chmod(path / 'cdwrapper', 0o744)
        cmd = str(path / 'cdwrapper')
    subprocess.call(
        [cmd], stdin=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    database.store_local_file(model, (path / basepath).with_suffix(model.filename_extension))
    database.store_local_file(model, (path / basepath).with_suffix('.lst'))
    database.store_local_file(model, (path / basepath).with_suffix('.ext'))
    database.store_local_file(model, (path / basepath).with_suffix('.phi'))
    database.store_local_file(model, (path / basepath).with_suffix('.cov'))
    database.store_local_file(model, (path / basepath).with_suffix('.cor'))
    database.store_local_file(model, (path / basepath).with_suffix('.coi'))
    for rec in model.control_stream.get_records('TABLE'):
        database.store_local_file(model, path / rec.path)
    # Read in results for the server side
    model.read_modelfit_results()
    # FIXME: the database path is changed in write
    model.database = database

    return model


def nmfe_path():
    if os.name == 'nt':
        nmfe_candidates = ['nmfe74.bat', 'nmfe75.bat', 'nmfe73.bat']
    else:
        nmfe_candidates = ['nmfe74', 'nmfe75', 'nmfe73']
    path = conf.default_nonmem_path
    if path != Path(''):
        path /= 'run'
    for nmfe in nmfe_candidates:
        candidate_path = path / nmfe
        if candidate_path.is_file():
            path = candidate_path
            break
    else:
        raise FileNotFoundError(f'Cannot find nmfe script for NONMEM ({path})')
    return str(path)



# See issue #20
# extensions = ['altair.sphinxext.altairplot']
extensions = ['pharmpy.reporting.altairplot']
html_sidebars = {
    '**': ['localtoc.html'],
}
html_theme_options = {'page_width': '1750px'}
html_title = 'Results'
master_doc = 'results'



from pathlib import Path

from .option_record import OptionRecord


class EtasRecord(OptionRecord):
    @property
    def path(self):
        file_option = self.option_pairs['FILE']
        return Path(file_option)

    @path.setter
    def path(self, value):
        self.set_option('FILE', value)



import re

import sympy
from sympy import Derivative, Eq, Function, Piecewise

from pharmpy.model import ModelSyntaxError
from pharmpy.statements import Assignment, Bolus, CompartmentalSystem, ExplicitODESystem, Infusion
from pharmpy.symbols import symbol


def compartmental_model(model, advan, trans, des=None):
    if advan == 'ADVAN1':
        cm = CompartmentalSystem()
        central = cm.add_compartment('CENTRAL')
        output = cm.add_compartment('OUTPUT')
        cm.add_flow(central, output, _advan1and2_trans(trans))
        dose = _dosing(model, 1)
        central.dose = dose
        central.lag_time = get_alag(model, 1)
        ass = _f_link_assignment(model, central)
        comp_map = {'CENTRAL': 1, 'OUTPUT': 2}
    elif advan == 'ADVAN2':
        cm = CompartmentalSystem()
        depot = cm.add_compartment('DEPOT')
        central = cm.add_compartment('CENTRAL')
        output = cm.add_compartment('OUTPUT')
        cm.add_flow(central, output, _advan1and2_trans(trans))
        cm.add_flow(depot, central, symbol('KA'))
        dose = _dosing(model, 1)
        depot.dose = dose
        depot.lag_time = get_alag(model, 1)
        central.lag_time = get_alag(model, 2)
        ass = _f_link_assignment(model, central)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'OUTPUT': 3}
    elif advan == 'ADVAN3':
        cm = CompartmentalSystem()
        central = cm.add_compartment('CENTRAL')
        peripheral = cm.add_compartment('PERIPHERAL')
        output = cm.add_compartment('OUTPUT')
        k, k12, k21 = _advan3_trans(trans)
        cm.add_flow(central, output, k)
        cm.add_flow(central, peripheral, k12)
        cm.add_flow(peripheral, central, k21)
        dose = _dosing(model, 1)
        central.dose = dose
        central.lag_time = get_alag(model, 1)
        peripheral.lag_time = get_alag(model, 2)
        ass = _f_link_assignment(model, central)
        comp_map = {'CENTRAL': 1, 'PERIPHERAL': 2, 'OUTPUT': 3}
    elif advan == 'ADVAN4':
        cm = CompartmentalSystem()
        depot = cm.add_compartment('DEPOT')
        central = cm.add_compartment('CENTRAL')
        peripheral = cm.add_compartment('PERIPHERAL')
        output = cm.add_compartment('OUTPUT')
        k, k23, k32, ka = _advan4_trans(trans)
        cm.add_flow(depot, central, ka)
        cm.add_flow(central, output, k)
        cm.add_flow(central, peripheral, k23)
        cm.add_flow(peripheral, central, k32)
        dose = _dosing(model, 1)
        depot.dose = dose
        depot.lag_time = get_alag(model, 1)
        central.lag_time = get_alag(model, 2)
        peripheral.lag_time = get_alag(model, 3)
        ass = _f_link_assignment(model, central)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'PERIPHERAL': 3, 'OUTPUT': 4}
    elif advan == 'ADVAN5' or advan == 'ADVAN7':
        cm = CompartmentalSystem()
        modrec = model.control_stream.get_records('MODEL')[0]
        defobs = None
        defdose = None
        central = None
        depot = None
        first_dose = None
        compartments = []
        for i, (name, opts) in enumerate(modrec.compartments()):
            comp = cm.add_compartment(name)
            if 'DEFOBSERVATION' in opts:
                defobs = comp
            if 'DEFDOSE' in opts:
                defdose = comp
                dose_no = i + 1
            if name == 'CENTRAL':
                central = comp
            elif name == 'DEPOT':
                depot = comp
                depot_no = i + 1
            if first_dose is None and 'NODOSE' not in opts:
                first_dose = comp
                first_dose_no = i + 1
            compartments.append(comp)
        output = cm.add_compartment('OUTPUT')
        compartments.append(output)
        comp_map = {comp.name: i + 1 for i, comp in enumerate(compartments)}
        ncomp = i + 2
        if not defobs:
            if central:
                defobs = central
            else:
                defobs = compartments[0]
        if not defdose:
            if depot:
                defdose = depot
                dose_no = depot_no
            elif first_dose is not None:
                defdose = first_dose
                dose_no = first_dose_no
            else:
                raise ModelSyntaxError('Dosing compartment is unknown')
        for from_n, to_n, rate in _find_rates(model, ncomp):
            cm.add_flow(compartments[from_n - 1], compartments[to_n - 1], rate)
        dose = _dosing(model, dose_no)
        defdose.dose = dose
        for i, comp in enumerate(compartments):
            if i == len(compartments) - 1:
                break
            comp.lag_time = get_alag(model, i)
        ass = _f_link_assignment(model, defobs)
    elif advan == 'ADVAN10':
        cm = CompartmentalSystem()
        central = cm.add_compartment('CENTRAL')
        output = cm.add_compartment('OUTPUT')
        vm = symbol('VM')
        km = symbol('KM')
        dose = _dosing(model, 1)
        central.dose = dose
        t = symbol('t')
        cm.add_flow(central, output, vm / (km + sympy.Function(central.amount.name)(t)))
        central.lag_time = get_alag(model, 1)
        ass = _f_link_assignment(model, central)
        comp_map = {'CENTRAL': 1, 'OUTPUT': 2}
    elif advan == 'ADVAN11':
        cm = CompartmentalSystem()
        central = cm.add_compartment('CENTRAL')
        per1 = cm.add_compartment('PERIPHERAL1')
        per2 = cm.add_compartment('PERIPHERAL2')
        output = cm.add_compartment('OUTPUT')
        k, k12, k21, k13, k31 = _advan11_trans(trans)
        cm.add_flow(central, output, k)
        cm.add_flow(central, per1, k12)
        cm.add_flow(per1, central, k21)
        cm.add_flow(central, per2, k13)
        cm.add_flow(per2, central, k31)
        dose = _dosing(model, 1)
        central.dose = dose
        central.lag_time = get_alag(model, 1)
        per1.lag_time = get_alag(model, 2)
        per2.lag_time = get_alag(model, 3)
        ass = _f_link_assignment(model, central)
        comp_map = {'CENTRAL': 1, 'PERIPHERAL1': 2, 'PERIPHERAL2': 3, 'OUTPUT': 4}
    elif advan == 'ADVAN12':
        cm = CompartmentalSystem()
        depot = cm.add_compartment('DEPOT')
        central = cm.add_compartment('CENTRAL')
        per1 = cm.add_compartment('PERIPHERAL1')
        per2 = cm.add_compartment('PERIPHERAL2')
        output = cm.add_compartment('OUTPUT')
        k, k23, k32, k24, k42, ka = _advan12_trans(trans)
        cm.add_flow(depot, central, ka)
        cm.add_flow(central, output, k)
        cm.add_flow(central, per1, k23)
        cm.add_flow(per1, central, k32)
        cm.add_flow(central, per2, k24)
        cm.add_flow(per2, central, k42)
        dose = _dosing(model, 1)
        depot.dose = dose
        depot.lag_time = get_alag(model, 1)
        central.lag_time = get_alag(model, 2)
        per1.lag_time = get_alag(model, 3)
        per2.lag_time = get_alag(model, 4)
        ass = _f_link_assignment(model, central)
        comp_map = {'DEPOT': 1, 'CENTRAL': 2, 'PERIPHERAL1': 3, 'PERIPHERAL2': 4, 'OUTPUT': 5}
    elif des and all(str(s.symbol).startswith('DADT') for s in des.statements):
        rec_model = model.control_stream.get_records('MODEL')[0]

        subs_dict, comp_names = dict(), dict()
        comps = [c for c, _ in rec_model.compartments()]

        t = symbol('t')
        for i, c in enumerate(comps, 1):
            a = Function(f'A_{c}')
            subs_dict[f'DADT({i})'] = Derivative(a(t))
            subs_dict[f'A({i})'] = a(t)
            comp_names[f'A({i})'] = a

        sset = des.statements
        sset.subs(subs_dict)

        a_out = Function('A_OUTPUT')
        dose = _dosing(model, 1)

        ics = {v(0): sympy.Integer(0) for v in comp_names.values()}
        ics[a_out(0)] = sympy.Integer(0)
        ics[comp_names['A(1)'](0)] = dose.amount

        dadt_dose = sset.find_assignment(str(subs_dict['DADT(1)']))

        if len(comps) > 1:
            dadt_rest = [Eq(s.symbol, s.expression) for s in sset if s != dadt_dose]
            lhs_sum = dadt_dose.expression
            for eq in dadt_rest:
                lhs_sum += eq.rhs
            dadt_out = Eq(Derivative(a_out(t)), -lhs_sum)
            dadt_rest.append(sympy.simplify(dadt_out))
        else:
            dadt_rest = [Eq(Derivative(a_out(t)), dadt_dose.expression * -1)]

        if isinstance(dose, Infusion):
            if dose.duration:
                rate = dose.amount / dose.duration
                duration = dose.duration
            else:
                rate = dose.rate
                duration = dose.amount / dose.rate

            dadt_dose.expression += Piecewise((rate, duration > t), (0, True))
            ics[comp_names['A(1)'](0)] = sympy.Integer(0)

        eqs = [Eq(dadt_dose.symbol, dadt_dose.expression)] + dadt_rest

        ode = ExplicitODESystem(eqs, ics)
        ass = _f_link_assignment(model, symbol('A_CENTRAL'))
        return ode, ass
    else:
        return None
    model._compartment_map = comp_map
    return cm, ass


def _f_link_assignment(model, compartment):
    f = symbol('F')
    try:
        fexpr = compartment.amount
    except AttributeError:
        fexpr = compartment
    pkrec = model.control_stream.get_records('PK')[0]
    if pkrec.statements.find_assignment('S1'):
        fexpr = fexpr / symbol('S1')
    ass = Assignment(f, fexpr)
    return ass


def _find_rates(model, ncomps):
    pkrec = model.control_stream.get_records('PK')[0]
    for stat in pkrec.statements:
        if hasattr(stat, 'symbol'):
            name = stat.symbol.name
            m = re.match(r'^K(\d+)(T\d+)?$', name)
            if m:
                if m.group(2):
                    from_n = int(m.group(1))
                    to_n = int(m.group(2)[1:])
                else:
                    n = m.group(1)
                    if len(n) == 2:
                        from_n = int(n[0])
                        to_n = int(n[1])
                    elif len(n) == 3:
                        f1 = int(n[0])
                        t1 = int(n[1:])
                        f2 = int(n[0:2])
                        t2 = int(n[2:])
                        q1 = f1 <= ncomps and t1 <= ncomps and t1 != 0
                        q2 = f2 <= ncomps and t2 <= ncomps
                        if q1 and q2:
                            raise ModelSyntaxError(
                                f'Rate parameter {n} is ambiguous. ' f'Use the KiTj notation.'
                            )
                        if q1:
                            from_n = f1
                            to_n = t1
                        elif q2:
                            from_n = f2
                            to_n = t2
                        else:
                            # Too large to or from compartment index. What would NONMEM do?
                            # Could also be too large later
                            continue
                    elif len(n) == 4:
                        from_n = int(n[0:2])
                        to_n = int(n[2:])
                if to_n == 0:
                    to_n = ncomps
                yield from_n, to_n, symbol(name)


def _advan1and2_trans(trans):
    if trans == 'TRANS2':
        return symbol('CL') / symbol('V')
    else:  # TRANS1 which is also the default
        return symbol('K')


def _advan3_trans(trans):
    if trans == 'TRANS3':
        return (
            symbol('CL') / symbol('V'),
            symbol('Q') / symbol('V'),
            symbol('Q') / (symbol('VSS') - symbol('V')),
        )
    elif trans == 'TRANS4':
        return (symbol('CL') / symbol('V1'), symbol('Q') / symbol('V1'), symbol('Q') / symbol('V2'))
    elif trans == 'TRANS5':
        return (
            symbol('ALPHA') * symbol('BETA') / symbol('K21'),
            symbol('ALPHA') + symbol('BETA') - symbol('K21') - symbol('K'),
            (symbol('AOB') * symbol('BETA') + symbol('ALPHA')) / (symbol('AOB') + 1),
        )
    elif trans == 'TRANS6':
        return (
            symbol('ALPHA') * symbol('BETA') / symbol('K21'),
            symbol('ALPHA') + symbol('BETA') - symbol('K21') - symbol('K'),
            symbol('K21'),
        )
    else:
        return (symbol('K'), symbol('K12'), symbol('K21'))


def _advan4_trans(trans):
    if trans == 'TRANS3':
        return (
            symbol('CL') / symbol('V'),
            symbol('Q') / symbol('V'),
            symbol('Q') / (symbol('VSS') - symbol('V')),
            symbol('KA'),
        )
    elif trans == 'TRANS4':
        return (
            symbol('CL') / symbol('V2'),
            symbol('Q') / symbol('V2'),
            symbol('Q') / symbol('V3'),
            symbol('KA'),
        )
    elif trans == 'TRANS5':
        return (
            symbol('ALPHA') * symbol('BETA') / symbol('K32'),
            symbol('ALPHA') + symbol('BETA') - symbol('K32') - symbol('K'),
            (symbol('AOB') * symbol('BETA') + symbol('ALPHA')) / (symbol('AOB') + 1),
            symbol('KA'),
        )
    elif trans == 'TRANS6':
        return (
            symbol('ALPHA') * symbol('BETA') / symbol('K32'),
            symbol('ALPHA') + symbol('BETA') - symbol('K32') - symbol('K'),
            symbol('K32'),
            symbol('KA'),
        )
    else:
        return (symbol('K'), symbol('K23'), symbol('K32'), symbol('KA'))


def _advan11_trans(trans):
    if trans == 'TRANS4':
        return (
            symbol('CL') / symbol('V1'),
            symbol('Q2') / symbol('V1'),
            symbol('Q2') / symbol('V2'),
            symbol('Q3') / symbol('V1'),
            symbol('Q3') / symbol('V3'),
        )
    elif trans == 'TRANS6':
        return (
            symbol('ALPHA') * symbol('BETA') * symbol('GAMMA') / (symbol('K21') * symbol('K31')),
            symbol('ALPHA')
            + symbol('BETA')
            + symbol('GAMMA')
            - symbol('K')
            - symbol('K13')
            - symbol('K21')
            - symbol('K31'),
            symbol('K21'),
            (
                symbol('ALPHA') * symbol('BETA')
                + symbol('ALPHA') * symbol('GAMMA')
                + symbol('BETA') * symbol('GAMMA')
                + symbol('K31') * symbol('K31')
                - symbol('K31') * (symbol('ALPHA') + symbol('BETA') + symbol('GAMMA'))
                - symbol('K') * symbol('K21')
            )
            / (symbol('K21') - symbol('K31')),
            symbol('K31'),
        )
    else:
        return (symbol('K'), symbol('K12'), symbol('K21'), symbol('K13'), symbol('K31'))


def _advan12_trans(trans):
    if trans == 'TRANS4':
        return (
            symbol('CL') / symbol('V2'),
            symbol('Q3') / symbol('V2'),
            symbol('Q3') / symbol('V3'),
            symbol('Q4') / symbol('V2'),
            symbol('Q4') / symbol('V4'),
            symbol('KA'),
        )
    elif trans == 'TRANS6':
        return (
            symbol('ALPHA') * symbol('BETA') * symbol('GAMMA') / (symbol('K32') * symbol('K42')),
            symbol('ALPHA')
            + symbol('BETA')
            + symbol('GAMMA')
            - symbol('K')
            - symbol('K24')
            - symbol('K32')
            - symbol('K42'),
            symbol('K32'),
            (
                symbol('ALPHA') * symbol('BETA')
                + symbol('ALPHA') * symbol('GAMMA')
                + symbol('BETA') * symbol('GAMMA')
                + symbol('K42') * symbol('K42')
                - symbol('K42') * (symbol('ALPHA') + symbol('BETA') + symbol('GAMMA'))
                - symbol('K') * symbol('K32')
            )
            / (symbol('K32') - symbol('K42')),
            symbol('K42'),
            symbol('KA'),
        )
    else:
        return (
            symbol('K'),
            symbol('K23'),
            symbol('K32'),
            symbol('K24'),
            symbol('K42'),
            symbol('KA'),
        )


def _dosing(model, dose_comp):
    colnames, _, _, _ = model._column_info()
    if 'RATE' in colnames:
        df = model.dataset
        if (df['RATE'] == 0).all():
            return Bolus('AMT')
        elif (df['RATE'] == -1).any():
            return Infusion('AMT', rate=f'R{dose_comp}')
        elif (df['RATE'] == -2).any():
            return Infusion('AMT', duration=f'D{dose_comp}')
        else:
            return Infusion('AMT', rate='RATE')
    else:
        return Bolus('AMT')


def get_alag(model, n):
    """Check if ALAGn is defined in model and return it else return 0"""
    alag = f'ALAG{n}'
    pkrec = model.control_stream.get_records('PK')[0]
    if pkrec.statements.find_assignment(alag):
        return alag
    else:
        return 0



import numpy as np
import pandas as pd
import pytest

from pharmpy import Model
from pharmpy.modeling import (
    create_rng,
    load_example_model,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    sample_parameters_uniformly,
)


def test_create_rng():
    rng = create_rng(23)
    assert rng.standard_normal() == 0.5532605888887387

    rng = create_rng(23.0)
    assert rng.standard_normal() == 0.5532605888887387


def test_sample_parameters_uniformly():
    model = load_example_model("pheno")
    rng = create_rng(23)
    df = sample_parameters_uniformly(model, n=3, rng=rng)
    assert df['THETA(1)'][0] == 0.004877674495376137


def test_sample_parameter_from_covariance_matrix(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    rng = np.random.default_rng(318)
    samples = sample_parameters_from_covariance_matrix(model, n=3, rng=rng)
    correct = pd.DataFrame(
        {
            'THETA(1)': [0.004489330033579095, 0.004866193232279955, 0.004619661658761273],
            'THETA(2)': [0.9720563045663096, 1.0217868717352445, 0.9662036500731115],
            'THETA(3)': [0.19927467608338267, 0.237140948854298, 0.1979609848931148],
            'OMEGA(1,1)': [0.012012933626520568, 0.03859989956899462, 0.03228178862778379],
            'OMEGA(2,2)': [0.03718187653238525, 0.036766142234483934, 0.02433717922068797],
            'SIGMA(1,1)': [0.00962550646345379, 0.01311348785596405, 0.014054031573722888],
        }
    )
    pd.testing.assert_frame_equal(samples, correct, atol=1e-6)
    # Make cov matrix non-posdef
    model.modelfit_results.covariance_matrix['THETA(1)']['THETA(1)'] = -1
    with pytest.warns(UserWarning):
        sample_parameters_from_covariance_matrix(model, n=1, force_posdef_covmatrix=True)


def test_sample_individual_estimates(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    rng = np.random.default_rng(86)
    samples = sample_individual_estimates(model, rng=rng)
    assert len(samples) == 59 * 100
    assert list(samples.columns) == ['ETA(1)', 'ETA(2)']
    assert pytest.approx(samples.iloc[0]['ETA(1)'], 1e-5) == 0.21179186940672637
    assert pytest.approx(samples.iloc[0]['ETA(2)'], 1e-5) == -0.05771736555248238

    restricted = sample_individual_estimates(
        model, parameters=['ETA(2)'], samples_per_id=1, rng=rng
    )
    assert len(restricted) == 59
    assert restricted.columns == ['ETA(2)']
    assert pytest.approx(restricted.iloc[0]['ETA(2)'], 1e-5) == 0.06399039578129821



"""
NONMEM simulation record class.
"""


from .record import Record


class SimulationRecord(Record):
    @property
    def nsubs(self):
        """Number of subproblems"""
        n = self.root.nsubs.INT
        if n == 0:
            # According to NONMEM documentation 0 means 1 here
            n = 1
        return n



"""
:meta private:
"""

from pharmpy.modeling.help_functions import _format_input_list, _get_etas


def remove_iiv(model, to_remove=None):
    """
    Removes all IIV etas given a list with eta names and/or parameter names.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    to_remove : str, list
        Name/names of etas and/or name/names of individual parameters to remove.
        If None, all etas that are IIVs will be removed. None is default.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_iiv(model)  # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("CL")
    CL := TVCL

    >>> model = load_example_model("pheno")
    >>> remove_iiv(model, "V")  # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("V")
    V := TVV

    See also
    --------
    remove_iov
    add_iiv
    add_iov

    """
    rvs, sset = model.random_variables, model.statements
    to_remove = _format_input_list(to_remove)
    etas = _get_etas(model, to_remove, include_symbols=True)

    for eta in etas:
        sset.subs({eta.symbol: 0})
        del rvs[eta]

    model.random_variables = rvs
    model.statements = sset

    model.modelfit_results = None

    return model



import math
import re
import warnings

import numpy as np
import sympy
import sympy.stats

import pharmpy.math
from pharmpy.model import ModelSyntaxError
from pharmpy.parameter import Parameter, Parameters
from pharmpy.parse_utils.generic import (
    AttrToken,
    AttrTree,
    insert_after,
    insert_before_or_at_end,
    remove_token_and_space,
)
from pharmpy.random_variables import RandomVariable, RandomVariables
from pharmpy.symbols import symbol

from .parsers import OmegaRecordParser
from .record import Record


class OmegaRecord(Record):
    def parameters(self, start_omega, previous_size, seen_labels=None):
        """Get a Parameters for this omega record"""
        row = start_omega
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        same = bool(self.root.find('same'))
        parameters = Parameters()
        coords = []

        try:
            self.comment_map
        except AttributeError:
            self.comment_map = dict()

        if seen_labels is None:
            seen_labels = set()
        if not (block or bare_block):
            for node in self.root.all('diag_item'):
                init = node.init.NUMERIC
                fixed = bool(node.find('FIX'))
                sd = bool(node.find('SD'))
                var = bool(node.find('VAR'))
                n = node.n.INT if node.find('n') else 1
                if sd and var:
                    raise ModelSyntaxError(
                        f'Initial estimate for {self.name.upper()} cannot be both'
                        f' on SD and VAR scale\n{self.root}'
                    )
                if init == 0 and not fixed:
                    raise ModelSyntaxError(
                        f'If initial estimate for {self.name.upper()} is 0 it'
                        f' must be set to FIX'
                    )
                if sd:
                    init = init ** 2
                for _ in range(n):
                    name = self._find_label(node, seen_labels)
                    comment = self._get_name(node)
                    if not name:
                        name = f'{self.name}({row},{row})'
                    if comment:
                        self.comment_map[name] = comment
                    seen_labels.add(name)
                    coords.append((row, row))
                    param = Parameter(name, init, lower=0, fix=fixed)
                    parameters.append(param)
                    row += 1
            size = 1
            next_omega = row
        else:
            inits = []
            if bare_block:
                size = previous_size
            else:
                size = self.root.block.size.INT
            fix, sd, corr, cholesky = self._block_flags()
            labels = []
            for node in self.root.all('omega'):
                init = node.init.NUMERIC
                n = node.n.INT if node.find('n') else 1
                inits += [init] * n
                name = self._find_label(node, seen_labels)
                if name is not None:
                    seen_labels.add(name)
                labels.append(name)
                if n > 1:
                    labels.extend([None] * (n - 1))
            if not same:
                if size != pharmpy.math.triangular_root(len(inits)):
                    raise ModelSyntaxError('Wrong number of inits in BLOCK')
                if not cholesky:
                    A = pharmpy.math.flattened_to_symmetric(inits)
                    if corr:
                        for i in range(size):
                            for j in range(size):
                                if i != j:
                                    if sd:
                                        A[i, j] = A[i, i] * A[j, j] * A[i, j]
                                    else:
                                        A[i, j] = math.sqrt(A[i, i]) * math.sqrt(A[j, j]) * A[i, j]
                    if sd:
                        np.fill_diagonal(A, A.diagonal() ** 2)
                else:
                    L = np.zeros((size, size))
                    inds = np.tril_indices_from(L)
                    L[inds] = inits
                    A = L @ L.T
                label_index = 0
                for i in range(size):
                    for j in range(0, i + 1):
                        name = labels[label_index]
                        if name is None:
                            name = f'{self.name}({i + start_omega},{j + start_omega})'
                        coords.append((i + start_omega, j + start_omega))
                        init = A[i, j]
                        lower = None if i != j else 0
                        param = Parameter(name, init, lower=lower, fix=fix)
                        parameters.append(param)
                        label_index += 1
            next_omega = start_omega + size
        try:
            self.name_map
        except AttributeError:
            self.name_map = {name: c for i, (name, c) in enumerate(zip(parameters.names, coords))}
        return parameters, next_omega, size

    def _find_label(self, node, seen_labels):
        """Find label from comment of omega parameter"""
        # needed to avoid circular import with Python 3.6
        import pharmpy.plugins.nonmem as nonmem

        name = None
        if 'comment' in nonmem.conf.parameter_names:
            name = self._get_name(node)
            if name in seen_labels:
                warnings.warn(
                    f'The parameter name {name} is duplicated. Falling back to basic NONMEM names.'
                )
                name = None
        return name

    def _get_name(self, node):
        name = None
        found = False
        for subnode in self.root.tree_walk():
            if id(subnode) == id(node):
                found = True
                continue
            if found and (subnode.rule == 'omega' or subnode.rule == 'diag_item'):
                break
            if found and (subnode.rule == 'NEWLINE' or subnode.rule == 'COMMENT'):
                m = re.search(r';\s*([a-zA-Z_]\w*)', str(subnode))
                if m:
                    name = m.group(1)
                    break
        return name

    def _block_flags(self):
        """Get a tuple of all interesting flags for block"""
        fix = bool(self.root.find('FIX'))
        var = bool(self.root.find('VAR'))
        sd = bool(self.root.find('SD'))
        cov = bool(self.root.find('COV'))
        corr = bool(self.root.find('CORR'))
        cholesky = bool(self.root.find('CHOLESKY'))
        for node in self.root.all('omega'):
            if node.find('FIX'):
                if fix:
                    raise ModelSyntaxError('Cannot specify option FIX more than once')
                else:
                    fix = True
            if node.find('VAR'):
                if var or sd or cholesky:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    var = True
            if node.find('SD'):
                if sd or var or cholesky:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    sd = True
            if node.find('COV'):
                if cov or corr:
                    raise ModelSyntaxError(
                        'Cannot specify either option COVARIANCE or ' 'CORRELATION more than once'
                    )
                else:
                    cov = True
            if node.find('CORR'):
                if corr or cov:
                    raise ModelSyntaxError(
                        'Cannot specify either option COVARIANCE or ' 'CORRELATION more than once'
                    )
                else:
                    corr = True
            if node.find('CHOLESKY'):
                if cholesky or var or sd:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    cholesky = True
        return fix, sd, corr, cholesky

    def _rv_name(self, num):
        if self.name == 'OMEGA':
            rv_name = 'ETA'
        else:
            rv_name = 'EPS'
        return f'{rv_name}({num})'

    def _rv_vector_name(self, omega_numbers):
        rv_strs = []
        for om in omega_numbers:
            name = self._rv_name(om)
            rv_strs.append(name)
        return '(' + ', '.join(rv_strs) + ')'

    def _get_param_name(self, row, col):
        """Get the name of the current OMEGA/SIGMA at (row, col)"""
        invmap = {value: key for key, value in self.name_map.items()}
        name = invmap[(row, col)]
        return name

    def update(self, parameters, first_omega, previous_size):
        """From a Parameters update the OMEGAs in this record
        returns the next omega number
        """
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        if not (block or bare_block):
            size = 1
            i = first_omega
            new_nodes = []
            for node in self.root.children:
                if node.rule != 'diag_item':
                    new_nodes.append(node)
                else:
                    sd = bool(node.find('SD'))
                    fix = bool(node.find('FIX'))
                    n = node.n.INT if node.find('n') else 1
                    new_inits = []
                    new_fix = []
                    for j in range(i, i + n):
                        name = self._get_param_name(j, j)
                        if not sd:
                            value = float(parameters[name].init)
                        else:
                            value = parameters[name].init ** 0.5
                        value = int(value) if value.is_integer() else value
                        new_inits.append(value)
                        new_fix.append(parameters[name].fix)
                    if n == 1 or (
                        new_inits.count(new_inits[0]) == len(new_inits)
                        and new_fix.count(new_fix[0]) == len(new_fix)
                    ):  # All equal?
                        if float(str(node.init)) != new_inits[0]:
                            node.init.tokens[0].value = str(new_inits[0])
                        if new_fix[0] != fix:
                            if new_fix[0]:
                                insert_before_or_at_end(
                                    node, 'RPAR', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                                )
                            else:
                                remove_token_and_space(node, 'FIX')
                        new_nodes.append(node)
                    else:
                        # Need to split xn
                        new_children = []
                        for child in node.children:
                            if child.rule != 'n':
                                new_children.append(child)
                        node.children = new_children
                        for j, init in enumerate(new_inits):
                            new_node = AttrTree.transform(node)
                            if float(str(new_node.init)) != init:
                                new_node.init.tokens[0].value = str(init)
                            if new_fix[j] != fix:
                                insert_before_or_at_end(
                                    node, 'RPAR', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                                )
                            else:
                                remove_token_and_space(node, 'FIX')
                            new_nodes.append(new_node)
                            if j != len(new_inits) - 1:  # Not the last
                                new_nodes.append(AttrTree.create('ws', {'WS': ' '}))
                    i += n
            self.root.children = new_nodes
            next_omega = i
        else:
            same = bool(self.root.find('same'))
            if same:
                return first_omega + previous_size, previous_size
            size = self.root.block.size.INT
            fix, sd, corr, cholesky = self._block_flags()
            inits = []
            new_fix = []
            for row in range(first_omega, first_omega + size):
                for col in range(first_omega, row + 1):
                    name = self._get_param_name(row, col)
                    inits.append(parameters[name].init)
                    new_fix.append(parameters[name].fix)
            if len(set(new_fix)) != 1:  # Not all true or all false
                raise ValueError('Cannot only fix some parameters in block')

            A = pharmpy.math.flattened_to_symmetric(inits)

            if corr:
                for i in range(size):
                    for j in range(size):
                        if i != j:
                            A[i, j] = A[i, j] / (math.sqrt(A[i, i]) * math.sqrt(A[j, j]))
            if sd:
                np.fill_diagonal(A, A.diagonal() ** 0.5)

            if cholesky:
                A = np.linalg.cholesky(A)

            inds = np.tril_indices_from(A)
            array = list(A[inds])
            i = 0
            new_nodes = []
            for node in self.root.children:
                if node.rule != 'omega':
                    new_nodes.append(node)
                else:
                    n = node.n.INT if node.find('n') else 1
                    if array[i : i + n].count(array[i]) == n:  # All equal?
                        if float(str(node.init)) != array[i]:
                            value = int(array[i]) if float(array[i]).is_integer() else array[i]
                            node.init.tokens[0].value = str(value)
                        new_nodes.append(node)
                    else:
                        # Need to split xn
                        new_children = []
                        for child in node.children:
                            if child.rule not in ['n', 'LPAR', 'RPAR']:
                                new_children.append(child)
                        node.children = new_children
                        for j, init in enumerate(array[i : i + n]):
                            new_node = AttrTree.transform(node)
                            if float(str(new_node.init)) != init:
                                init = int(init) if float(init).is_integer() else init
                                new_node.init.tokens[0].value = str(init)
                            new_nodes.append(new_node)
                            if j != n - 1:  # Not the last
                                new_nodes.append(AttrTree.create('ws', {'WS': ' '}))
                    i += n
            self.root.children = new_nodes
            if new_fix[0] != fix:
                if new_fix[0]:
                    insert_after(
                        self.root, 'block', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                    )
                else:
                    remove_token_and_space(self.root, 'FIX', recursive=True)
            next_omega = first_omega + size
        return next_omega, size

    def random_variables(self, start_omega, previous_cov=None):
        """Get a RandomVariableSet for this omega record

        start_omega - the first omega in this record
        previous_cov - the matrix of the previous omega block
        """
        same = bool(self.root.find('same'))
        if not hasattr(self, 'name_map') and not same:
            if isinstance(previous_cov, sympy.Symbol):
                prev_size = 1
            elif previous_cov is not None:
                prev_size = len(previous_cov)
            else:
                prev_size = None
            self.parameters(start_omega, prev_size)
        if hasattr(self, 'name_map'):
            rev_map = {value: key for key, value in self.name_map.items()}
        next_cov = None  # The cov matrix if a block
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        zero_fix = []
        etas = []
        if not (block or bare_block):
            rvs = RandomVariables()
            i = start_omega
            numetas = len(self.root.all('diag_item'))
            for node in self.root.all('diag_item'):
                init = node.init.NUMERIC
                fixed = bool(node.find('FIX'))
                name = self._rv_name(i)
                if not (init == 0 and fixed):  # 0 FIX are not RVs
                    eta = RandomVariable.normal(name, 'iiv', 0, symbol(rev_map[(i, i)]))
                    rvs.append(eta)
                    etas.append(eta.name)
                else:
                    zero_fix.append(name)
                    etas.append(name)
                i += 1
        else:
            if bare_block:
                numetas = previous_cov.rows
            else:
                numetas = self.root.block.size.INT
            params, _, _ = self.parameters(
                start_omega, previous_cov.rows if hasattr(previous_cov, 'rows') else None
            )
            all_zero_fix = True
            for param in params:
                if not (param.init == 0 and param.fix):
                    all_zero_fix = False
            if all_zero_fix and len(params) > 0 or (previous_cov == 'ZERO' and same):
                names = [self._rv_name(i) for i in range(start_omega, start_omega + numetas)]
                self.eta_map = {eta: start_omega + i for i, eta in enumerate(names)}
                return RandomVariables(), start_omega + numetas, 'ZERO', names
            if numetas > 1:
                names = [self._rv_name(i) for i in range(start_omega, start_omega + numetas)]
                means = [0] * numetas
                if same:
                    rvs = RandomVariable.joint_normal(names, 'iiv', means, previous_cov)
                    etas = [rv.name for rv in rvs]
                    next_cov = previous_cov
                else:
                    cov = sympy.zeros(numetas)
                    for row in range(numetas):
                        for col in range(row + 1):
                            cov[row, col] = symbol(rev_map[(start_omega + row, start_omega + col)])
                            if row != col:
                                cov[col, row] = cov[row, col]
                    next_cov = cov
                    rvs = RandomVariable.joint_normal(names, 'iiv', means, cov)
                    etas = [rv.name for rv in rvs]
            else:
                rvs = RandomVariables()
                name = self._rv_name(start_omega)
                if same:
                    sym = previous_cov
                else:
                    sym = symbol(rev_map[(start_omega, start_omega)])
                eta = RandomVariable.normal(name, 'iiv', 0, sym)
                next_cov = sym
                rvs.append(eta)
                etas.append(eta.name)

        if self.name == 'OMEGA':
            if same:
                level = 'IOV'
            else:
                level = 'IIV'
        else:
            level = 'RUV'
        for rv in rvs:
            rv.level = level
        self.eta_map = {eta: start_omega + i for i, eta in enumerate(etas)}
        return rvs, start_omega + numetas, next_cov, zero_fix

    def renumber(self, new_start):
        old_start = min(self.eta_map.values())
        if new_start != old_start:
            for name in self.eta_map:
                self.eta_map[name] += new_start - old_start
            for name in self.name_map:
                old_row, old_col = self.name_map[name]
                self.name_map[name] = (
                    old_row + new_start - old_start,
                    old_col + new_start - old_start,
                )

    def remove(self, names):
        """Remove some etas from block given eta names"""
        first_omega = min(self.eta_map.values())
        indices = {self.eta_map[name] - first_omega for name in names}
        for name in names:
            del self.eta_map[name]

        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        same = bool(self.root.find('same'))
        if not (block or bare_block):
            keep = []
            i = 0
            for node in self.root.children:
                if node.rule == 'diag_item':
                    if i not in indices:
                        keep.append(node)
                    i += 1
                else:
                    keep.append(node)
            self.root.children = keep
        elif same and not bare_block:
            self.root.block.size.INT = len(self) - len(indices)
        elif block:
            fix, sd, corr, cholesky = self._block_flags()
            inits = []
            for node in self.root.all('omega'):
                init = node.init.NUMERIC
                n = node.n.INT if node.find('n') else 1
                inits += [init] * n
            A = pharmpy.math.flattened_to_symmetric(inits)
            A = np.delete(A, list(indices), axis=0)
            A = np.delete(A, list(indices), axis=1)
            s = f' BLOCK({len(A)})'
            if fix:
                s += ' FIX'
            if sd:
                s += ' SD'
            if corr:
                s += ' CORR'
            if cholesky:
                s += ' CHOLESKY'
            s += '\n'
            for row in range(0, len(A)):
                s += ' '.join(np.atleast_1d(A[row, 0 : (row + 1)]).astype(str)) + '\n'
            parser = OmegaRecordParser(s)
            self.root = parser.root

    def update_name_map(self, trans):
        """Update name_map given dict from -> to"""
        for key, value in trans.items():
            if key in self.name_map:
                n = self.name_map[key]
                del self.name_map[key]
                self.name_map[value] = n

    def __len__(self):
        return len(self.eta_map)



from pharmpy.tools.resmod.results import psn_resmod_results


def test_resmod_results(testdata):
    res = psn_resmod_results(testdata / 'psn' / 'resmod_dir1')
    assert list(res.models['dOFV']) == [
        -1.31,
        -3.34,
        -13.91,
        -18.54,
        -8.03,
        -4.20,
        -0.25,
        -1.17,
        -0.00,
        -0.09,
        -2.53,
        -3.12,
        -3.60,
        -25.62,
        -7.66,
        -0.03,
        -5.53,
    ]


def test_resmod_results_dvid(testdata):
    res = psn_resmod_results(testdata / 'psn' / 'resmod_dir2')
    df = res.models
    assert df['dOFV'].loc[1, '1', 'autocorrelation'] == -0.74
    assert df['dOFV'].loc[1, 'sum', 'tdist'] == -35.98



"""Pretty printing of :class:`generic.AttrTree` via :class:`treeprint.Node`."""
from functools import partial

import lark

from . import treeprint

MAXLEN = 60


def _preview(content):
    """Represents str and cut into shorts if too long"""
    cut = len(content) - MAXLEN
    if cut > 0:
        preview = '"' + repr(content[0:MAXLEN])[1:-1] + '"'
        preview = '%s+%d' % (preview, cut)
    else:
        preview = '"' + repr(content)[1:-1] + '"'
    return preview


def _formatter(content, ast_node):
    """Formats AST node (Tree and Token) for treeprint"""
    if not content:
        return '%s' % (ast_node.rule,)
    elif content != 'full':
        return '%s %s' % (ast_node.rule, _preview(str(ast_node)))
    lines = [ast_node.rule] + str(ast_node).splitlines(True)
    lines[1:] = [_preview(line) for line in lines[1:]]
    return '\n'.join(lines)


def _format_tree(content, ast_tree):
    return _formatter(content, ast_tree)


def _format_token(content, ast_token):
    return _formatter(content, ast_token)


def transform(ast_tree, content=True):
    """
    Traverses tree and generates :class:`treeprint.Node`, which can format a
    multiline (command) tree-styled string.

    Args:
        content: If True, include text preview (or all if 'full') on each node.

    Returns:
        Multiline string, ready for the printing press.
    """

    if not any(isinstance(ast_tree, x) for x in [lark.Tree, lark.lexer.Token]):
        raise TypeError(
            "can't transform %s object (is not a Lark Tree or Token')"
            % repr(ast_tree.__class__.__name__)
        )
    tree = treeprint.Node(ast_tree)
    try:
        nodes = ast_tree.children
    except AttributeError:
        tree.set_formatter(partial(_format_token, content))
    else:
        if isinstance(nodes, str):
            raise TypeError(
                "'children' of tree appears to be 'str' (expects list/iterable): %s"
                % repr(ast_tree)
            )
        tree.set_formatter(partial(_format_tree, content))
        for ast_node in list(nodes):
            node = transform(ast_node, content)
            tree.add(node)
    return tree



import shutil

import pytest

import pharmpy.modeling as modeling
from pharmpy import Model
from pharmpy.config import site_config_dir, user_config_dir
from pharmpy.plugins.nonmem import conf
from pharmpy.utils import TemporaryDirectoryChanger


def test_configuration():
    print("User config dir:", user_config_dir())
    print("Site config dir:", site_config_dir())
    print("Default NONMEM path:", conf.default_nonmem_path)
    assert (conf.default_nonmem_path / 'license' / 'nonmem.lic').is_file()


def test_fit_single(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model('pheno.mod')
        model.dataset_path = tmp_path / 'pheno.dta'
        modeling.fit(model)
        rundir = tmp_path / 'modelfit_dir1'
        assert model.modelfit_results.ofv == pytest.approx(730.8947268137308)
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 1


def test_fit_multiple(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno_1.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno_1.dta')
        model_1 = Model('pheno_1.mod')
        model_1.dataset_path = tmp_path / 'pheno_1.dta'
        model_1.update_source()
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno_2.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno_2.dta')
        model_2 = Model('pheno_2.mod')
        model_2.dataset_path = tmp_path / 'pheno_2.dta'
        model_2.update_source()
        modeling.fit([model_1, model_2])
        rundir = tmp_path / 'modelfit_dir1'
        assert model_1.modelfit_results.ofv == pytest.approx(730.8947268137308)
        assert model_2.modelfit_results.ofv == pytest.approx(730.8947268137308)
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 2


def test_fit_copy(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno.dta')

        model_1 = Model('pheno.mod')
        model_1.dataset_path = tmp_path / 'pheno.dta'
        modeling.fit(model_1)

        rundir_1 = tmp_path / 'modelfit_dir1'
        assert rundir_1.is_dir()
        assert len(list((rundir_1 / 'models').iterdir())) == 1

        model_2 = modeling.copy_model(model_1, 'pheno_copy')
        modeling.update_inits(model_2)
        modeling.fit(model_2)

        rundir_2 = tmp_path / 'modelfit_dir1'
        assert rundir_2.is_dir()
        assert len(list((rundir_2 / 'models').iterdir())) == 1

        assert model_1.modelfit_results.ofv != model_2.modelfit_results.ofv
        mod_files = [
            path for path in tmp_path.iterdir() if path.is_file() and path.suffix == '.mod'
        ]
        assert len(mod_files) == 2
